#!/usr/bin/env python
import typer
from rich import print
from typing_extensions import Annotated
import numpy as np
import segno
from segno import DataOverflowError, writers
from typing import List
from enum import Enum



class Direction(str, Enum):
    LEFT_TO_RIGHT = "LTR"
    TOP_TO_BOTTOM = "TTB"
    RIGHT_TO_LEFT = "RTL"

# 5x5 block of bits that can be used to place decoration next to barcode.
class VisualMarker:
    size = 5

    def __init__(self, bits):
        # true means 1 bit, false means 0 bit (eg: 1 means black, 0 means white)
        self.bits = bits

    marker_dict = {
        'A': bytearray([
            0x00, 0x01, 0x01, 0x01, 0x00,  # preformatted
            0x01, 0x00, 0x00, 0x00, 0x01,
            0x01, 0x01, 0x01, 0x01, 0x01,
            0x01, 0x00, 0x00, 0x00, 0x01,
            0x01, 0x00, 0x00, 0x00, 0x01,
        ]),
        'B': bytearray([
            0x01, 0x01, 0x01, 0x01, 0x00,  # preformatted
            0x01, 0x00, 0x00, 0x00, 0x01,
            0x01, 0x01, 0x01, 0x01, 0x00,
            0x01, 0x00, 0x00, 0x00, 0x01,
            0x01, 0x01, 0x01, 0x01, 0x00,
        ]),
        'C': bytearray([
            0x00, 0x01, 0x01, 0x01, 0x01,  # preformatted
            0x01, 0x00, 0x00, 0x00, 0x00,
            0x01, 0x00, 0x00, 0x00, 0x00,
            0x01, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x01, 0x01, 0x01, 0x01,
        ]),
        'D': bytearray([
            0x01, 0x01, 0x01, 0x01, 0x00,  # preformatted
            0x01, 0x00, 0x00, 0x00, 0x01,
            0x01, 0x00, 0x00, 0x00, 0x01,
            0x01, 0x00, 0x00, 0x00, 0x01,
            0x01, 0x01, 0x01, 0x01, 0x00,
        ]),
        'E': bytearray([
            0x01, 0x01, 0x01, 0x01, 0x01,  # preformatted
            0x01, 0x00, 0x00, 0x00, 0x00,
            0x01, 0x01, 0x01, 0x01, 0x01,
            0x01, 0x00, 0x00, 0x00, 0x00,
            0x01, 0x01, 0x01, 0x01, 0x01,
        ]),
        'F': bytearray([
            0x01, 0x01, 0x01, 0x01, 0x01,  # preformatted
            0x01, 0x00, 0x00, 0x00, 0x00,
            0x01, 0x01, 0x01, 0x01, 0x01,
            0x01, 0x00, 0x00, 0x00, 0x00,
            0x01, 0x00, 0x00, 0x00, 0x00,
        ]),
        'G': bytearray([
            0x00, 0x01, 0x01, 0x01, 0x01,  # preformatted
            0x01, 0x00, 0x00, 0x00, 0x00,
            0x01, 0x00, 0x01, 0x01, 0x00,
            0x01, 0x00, 0x00, 0x00, 0x01,
            0x00, 0x01, 0x01, 0x01, 0x01,
        ]),
        'H': bytearray([
            0x01, 0x00, 0x00, 0x00, 0x01,  # preformatted
            0x01, 0x00, 0x00, 0x00, 0x01,
            0x01, 0x01, 0x01, 0x01, 0x01,
            0x01, 0x00, 0x00, 0x00, 0x01,
            0x01, 0x00, 0x00, 0x00, 0x01,
        ]),
        'I': bytearray([
            0x00, 0x00, 0x01, 0x00, 0x00,  # preformatted
            0x00, 0x00, 0x01, 0x00, 0x00,
            0x00, 0x00, 0x01, 0x00, 0x00,
            0x00, 0x00, 0x01, 0x00, 0x00,
            0x00, 0x00, 0x01, 0x00, 0x00,
        ]),
        'J': bytearray([
            0x00, 0x00, 0x01, 0x00, 0x00,  # preformatted
            0x00, 0x00, 0x01, 0x00, 0x00,
            0x00, 0x00, 0x01, 0x00, 0x00,
            0x00, 0x00, 0x01, 0x00, 0x00,
            0x00, 0x01, 0x00, 0x00, 0x00,
        ]),
        'K': bytearray([
            0x01, 0x00, 0x00, 0x00, 0x01,  # preformatted
            0x01, 0x00, 0x00, 0x01, 0x00,
            0x01, 0x01, 0x01, 0x00, 0x00,
            0x01, 0x00, 0x00, 0x01, 0x00,
            0x01, 0x00, 0x00, 0x00, 0x01,
        ]),
        'L': bytearray([
            0x01, 0x00, 0x00, 0x00, 0x00,  # preformatted
            0x01, 0x00, 0x00, 0x00, 0x00,
            0x01, 0x00, 0x00, 0x00, 0x00,
            0x01, 0x00, 0x00, 0x00, 0x00,
            0x01, 0x01, 0x01, 0x01, 0x01,
        ]),
        'M': bytearray([
            0x01, 0x01, 0x00, 0x01, 0x01,  # preformatted
            0x01, 0x00, 0x01, 0x00, 0x01,
            0x01, 0x00, 0x00, 0x00, 0x01,
            0x01, 0x00, 0x00, 0x00, 0x01,
            0x01, 0x00, 0x00, 0x00, 0x01,
        ]),
        'N': bytearray([
            0x01, 0x00, 0x00, 0x00, 0x01,  # preformatted
            0x01, 0x01, 0x00, 0x00, 0x01,
            0x01, 0x00, 0x01, 0x00, 0x01,
            0x01, 0x00, 0x00, 0x01, 0x01,
            0x01, 0x00, 0x00, 0x00, 0x01,
        ]),
        'O': bytearray([
            0x00, 0x01, 0x01, 0x01, 0x00,  # preformatted
            0x01, 0x00, 0x00, 0x00, 0x01,
            0x01, 0x00, 0x00, 0x00, 0x01,
            0x01, 0x00, 0x00, 0x00, 0x01,
            0x00, 0x01, 0x01, 0x01, 0x00,
        ]),
        'P': bytearray([
            0x01, 0x01, 0x01, 0x01, 0x00,  # preformatted
            0x01, 0x00, 0x00, 0x00, 0x01,
            0x01, 0x01, 0x01, 0x01, 0x00,
            0x01, 0x00, 0x00, 0x00, 0x00,
            0x01, 0x00, 0x00, 0x00, 0x00,
        ]),
        'Q': bytearray([
            0x00, 0x01, 0x01, 0x01, 0x00,  # preformatted
            0x01, 0x00, 0x00, 0x00, 0x01,
            0x01, 0x00, 0x01, 0x00, 0x01,
            0x01, 0x00, 0x00, 0x01, 0x00,
            0x00, 0x01, 0x01, 0x00, 0x01,
        ]),
        'R': bytearray([
            0x01, 0x01, 0x01, 0x01, 0x00,  # preformatted
            0x01, 0x00, 0x00, 0x01, 0x00,
            0x01, 0x01, 0x01, 0x00, 0x00,
            0x01, 0x00, 0x00, 0x01, 0x00,
            0x01, 0x00, 0x00, 0x00, 0x01,
        ]),
        'S': bytearray([
            0x00, 0x01, 0x01, 0x01, 0x01,  # preformatted
            0x01, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x01, 0x01, 0x01, 0x00,
            0x00, 0x00, 0x00, 0x00, 0x01,
            0x01, 0x01, 0x01, 0x01, 0x00,
        ]),
        'T': bytearray([
            0x01, 0x01, 0x01, 0x01, 0x01,  # preformatted
            0x00, 0x00, 0x01, 0x00, 0x00,
            0x00, 0x00, 0x01, 0x00, 0x00,
            0x00, 0x00, 0x01, 0x00, 0x00,
            0x00, 0x00, 0x01, 0x00, 0x00,
        ]),
        'U': bytearray([
            0x01, 0x00, 0x00, 0x00, 0x01,  # preformatted
            0x01, 0x00, 0x00, 0x00, 0x01,
            0x01, 0x00, 0x00, 0x00, 0x01,
            0x01, 0x00, 0x00, 0x00, 0x01,
            0x00, 0x01, 0x01, 0x01, 0x00,
        ]),
        'V': bytearray([
            0x01, 0x00, 0x00, 0x00, 0x01,  # preformatted
            0x01, 0x00, 0x00, 0x00, 0x01,
            0x00, 0x01, 0x00, 0x01, 0x00,
            0x00, 0x01, 0x01, 0x01, 0x00,
            0x00, 0x00, 0x01, 0x00, 0x00,
        ]),
        'W': bytearray([
            0x01, 0x00, 0x00, 0x00, 0x01,  # preformatted
            0x01, 0x00, 0x01, 0x00, 0x01,
            0x01, 0x00, 0x01, 0x00, 0x01,
            0x00, 0x01, 0x01, 0x01, 0x00,
            0x00, 0x01, 0x00, 0x01, 0x00,
        ]),
        'X': bytearray([
            0x01, 0x00, 0x00, 0x00, 0x01,  # preformatted
            0x00, 0x01, 0x00, 0x01, 0x00,
            0x00, 0x00, 0x01, 0x00, 0x00,
            0x00, 0x01, 0x00, 0x01, 0x00,
            0x01, 0x00, 0x00, 0x00, 0x01,
        ]),
        'Y': bytearray([
            0x01, 0x00, 0x00, 0x00, 0x01,  # preformatted
            0x01, 0x00, 0x00, 0x00, 0x01,
            0x00, 0x01, 0x01, 0x01, 0x00,
            0x00, 0x00, 0x01, 0x00, 0x00,
            0x00, 0x00, 0x01, 0x00, 0x00,
        ]),
        'Z': bytearray([
            0x01, 0x01, 0x01, 0x01, 0x01,  # preformatted
            0x00, 0x00, 0x00, 0x01, 0x00,
            0x00, 0x00, 0x01, 0x00, 0x00,
            0x00, 0x01, 0x00, 0x00, 0x00,
            0x01, 0x01, 0x01, 0x01, 0x01,
        ]),
        '0': bytearray([
            0x00, 0x00, 0x01, 0x00, 0x00,  # preformatted
            0x00, 0x01, 0x00, 0x01, 0x00,
            0x00, 0x01, 0x00, 0x01, 0x00,
            0x00, 0x01, 0x00, 0x01, 0x00,
            0x00, 0x00, 0x01, 0x00, 0x00,
        ]),
        '1': bytearray([
            0x00, 0x00, 0x01, 0x00, 0x00,  # preformatted
            0x00, 0x01, 0x01, 0x00, 0x00,
            0x01, 0x00, 0x01, 0x00, 0x00,
            0x00, 0x00, 0x01, 0x00, 0x00,
            0x00, 0x00, 0x01, 0x00, 0x00,
        ]),
        '2': bytearray([
            0x00, 0x01, 0x01, 0x01, 0x00,  # preformatted
            0x01, 0x00, 0x00, 0x00, 0x01,
            0x00, 0x00, 0x01, 0x01, 0x00,
            0x00, 0x01, 0x00, 0x00, 0x00,
            0x01, 0x01, 0x01, 0x01, 0x01,
        ]),
        '3': bytearray([
            0x01, 0x01, 0x01, 0x01, 0x00,  # preformatted
            0x00, 0x00, 0x00, 0x00, 0x01,
            0x01, 0x01, 0x01, 0x01, 0x00,
            0x00, 0x00, 0x00, 0x00, 0x01,
            0x01, 0x01, 0x01, 0x01, 0x00,
        ]),
        '4': bytearray([
            0x01, 0x00, 0x00, 0x00, 0x00,  # preformatted
            0x01, 0x00, 0x00, 0x01, 0x00,
            0x01, 0x01, 0x01, 0x01, 0x01,
            0x00, 0x00, 0x00, 0x01, 0x00,
            0x00, 0x00, 0x00, 0x01, 0x00,
        ]),
        '5': bytearray([
            0x01, 0x01, 0x01, 0x01, 0x01,  # preformatted
            0x01, 0x00, 0x00, 0x00, 0x00,
            0x01, 0x01, 0x01, 0x01, 0x00,
            0x00, 0x00, 0x00, 0x00, 0x01,
            0x01, 0x01, 0x01, 0x01, 0x00,
        ]),
        '6': bytearray([
            0x00, 0x01, 0x01, 0x01, 0x00,  # preformatted
            0x01, 0x00, 0x00, 0x00, 0x00,
            0x01, 0x01, 0x01, 0x01, 0x00,
            0x01, 0x00, 0x00, 0x00, 0x01,
            0x00, 0x01, 0x01, 0x01, 0x00,
        ]),
        '7': bytearray([
            0x01, 0x01, 0x01, 0x01, 0x01,  # preformatted
            0x00, 0x00, 0x00, 0x01, 0x00,
            0x00, 0x00, 0x01, 0x00, 0x00,
            0x00, 0x01, 0x00, 0x00, 0x00,
            0x01, 0x00, 0x00, 0x00, 0x00,
        ]),
        '8': bytearray([
            0x00, 0x01, 0x01, 0x01, 0x00,  # preformatted
            0x01, 0x00, 0x00, 0x00, 0x01,
            0x00, 0x01, 0x01, 0x01, 0x00,
            0x01, 0x00, 0x00, 0x00, 0x01,
            0x00, 0x01, 0x01, 0x01, 0x00,
        ]),
        '9': bytearray([
            0x00, 0x01, 0x01, 0x01, 0x00,  # preformatted
            0x01, 0x00, 0x00, 0x00, 0x01,
            0x00, 0x01, 0x01, 0x01, 0x01,
            0x00, 0x00, 0x00, 0x00, 0x01,
            0x00, 0x01, 0x01, 0x01, 0x00,
        ]),
        '.': bytearray([
            0x00, 0x00, 0x00, 0x00, 0x00,  # preformatted
            0x00, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0x01, 0x00, 0x00,
            0x00, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0x00, 0x00, 0x00,
        ]),
        ',': bytearray([
            0x00, 0x00, 0x00, 0x00, 0x00,  # preformatted
            0x00, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0x01, 0x00, 0x00,
            0x00, 0x00, 0x01, 0x00, 0x00,
        ]),
        ':': bytearray([
            0x00, 0x00, 0x00, 0x00, 0x00,  # preformatted
            0x00, 0x00, 0x01, 0x00, 0x00,
            0x00, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0x01, 0x00, 0x00,
        ]),
        ';': bytearray([
            0x00, 0x00, 0x00, 0x00, 0x00,  # preformatted
            0x00, 0x00, 0x01, 0x00, 0x00,
            0x00, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0x01, 0x00, 0x00,
            0x00, 0x00, 0x01, 0x00, 0x00,
        ]),
        '-': bytearray([
            0x00, 0x00, 0x00, 0x00, 0x00,  # preformatted
            0x00, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x01, 0x01, 0x01, 0x00,
            0x00, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0x00, 0x00, 0x00,
        ])
    }
    
    def generate_from_text(chars: str, direction:Direction, size_in_modules: int):
        # fill up with "."
        squares_cnt = (size_in_modules + 1) // (VisualMarker.size+1)
        leftover = (size_in_modules + 1) % (VisualMarker.size+1)
        if len(chars) < squares_cnt:
            chars = chars + '.' * (squares_cnt - len(chars))
        result: List[bytearray] = []
        for char in chars[:squares_cnt]:
            if char in VisualMarker.marker_dict:
                marker = VisualMarker.marker_dict[char]
            else:
                marker = VisualMarker.marker_dict['.']
            result.append(marker)
        return VisualMarker.generate_from_squares(result, direction, leftover)

    def generate_from_squares(marker_squares: List[bytearray], direction:Direction, padding_at_end: int):
        # spacer is a column of zeros if RTL/LTR, or a row if TTB/BTT
        axis = 0 if direction == Direction.TOP_TO_BOTTOM else 1
        spacer_dim = (1, VisualMarker.size) if axis==0 else (VisualMarker.size, 1)
        spacer = np.zeros(spacer_dim, dtype=np.uint8)
        result = np.array([])
        for marker in marker_squares:
            if len(marker) != VisualMarker.size*VisualMarker.size:
                raise ValueError("All markers must be 5x5")

            np_marker = np.array(marker).reshape(VisualMarker.size, VisualMarker.size)
            if result.size == 0:
                # the first item doesn't need a spacer
                result = np_marker
            else:
                # append in direction specified
                if direction == Direction.RIGHT_TO_LEFT:
                    result = np.concatenate((np_marker, spacer, result), axis=1)
                else:
                    result = np.concatenate((result, spacer, np_marker), axis=axis)
        if(padding_at_end > 0):
            padding_dim = (padding_at_end, VisualMarker.size) if axis==0 else (VisualMarker.size, padding_at_end)
            padding= np.zeros(padding_dim, dtype=np.uint8)

            # append in direction specified
            if direction == Direction.RIGHT_TO_LEFT:
                result = np.concatenate((padding, result), axis=1)
            else:
                result = np.concatenate((result, padding), axis=axis)
        return result


app = typer.Typer()


def _generate_qr_with_markers(qr_str, text, title, direction): 
    if title:
    #try to use standard size 10. Go bigger if 10 does not fit the data
        try:
            qr = segno.make_qr(qr_str, error="L", version=10)
        except DataOverflowError:
            qr = segno.make_qr(qr_str, error="L")
    else: 
        qr = segno.make_qr(qr_str, error="L")
    
    if(qr.mode != "alphanumeric"):
        print("[bold yellow]Large QR:[/bold yellow] Provided URL is not alphanumeric!")
    block_count = len(qr.matrix)
    print(f"[bold]Size:[/bold] {block_count}")
    print(f"[bold]Version:[/bold] {qr.version}")
    print(f"[bold]Error Level:[/bold] {qr.error}")
    
    
    qr_matrix = np.array(qr.matrix)
    visual_marker = VisualMarker.generate_from_text(text.upper(), direction, block_count)
    if title:
        title_marker = VisualMarker.generate_from_text(title.upper(), direction, block_count)
    
    append_axis = 1 if direction == Direction.TOP_TO_BOTTOM else 0
    padding_dim = (4, block_count) if append_axis==0 else (block_count, qr.default_border_size)
    padding= np.zeros(padding_dim, dtype=np.uint8)
    if title:
        combined_matrix = np.concatenate((title_marker, padding, qr_matrix, padding, visual_marker), axis=append_axis)
    else:
        combined_matrix = np.concatenate((qr_matrix, padding, visual_marker), axis=append_axis)
        
    return combined_matrix

def save_qr_with_markers(url, text="PAC", title=None, direction = Direction.LEFT_TO_RIGHT, fmt='png', path='qr'):
    combined_matrix = _generate_qr_with_markers(url, text="PAC", title=None, direction = Direction.LEFT_TO_RIGHT)
    outfile = f'{path}.{fmt}'
    match fmt:
        case 'png':
            writers.write_png(combined_matrix, combined_matrix.shape[::-1], out=outfile, border=9)
        case 'svg':
            writers.write_svg(combined_matrix, combined_matrix.shape[::-1], out=outfile, border=9)
            



def main(url: Annotated[str, typer.Argument(help="The PAC-ID to be rendered as QR.")],
    outfile: Annotated[typer.FileBinaryWrite, typer.Option(help="The file the qr will be written to.")] = "qr.svg",
    text: Annotated[str, typer.Option(help="The text of the PAC decoration.")] = "PAC",
    direction: Annotated[Direction, typer.Option(help="The position/direction of the PAC decoration.")] = Direction.TOP_TO_BOTTOM):

    save_qr_with_markers(url, text=text, direction=direction, path=outfile) 
    


if __name__ == "__main__":
    typer.run(main)




import re
from pydantic import model_validator
from labfreed.labfreed_infrastructure import LabFREED_BaseModel, ValidationMsgLevel, _quote_texts
from labfreed.well_known_keys.labfreed.well_known_keys import WellKnownKeys


_hsegment_pattern = r"[A-Za-z0-9_\-\.~!$&'()+,:;=@]|%[0-9A-Fa-f]{2}"

class IDSegment(LabFREED_BaseModel):
    """ Represents an id segment of a PAC-ID. It can be a value or a key value pair.
    """
    key:str|None = None
    ''' The key of the segment. This is optional.'''
    value:str
    ''' The value of the segment. (mandatory)'''

    @model_validator(mode="after")
    def _validate_segment(self):
        key = self.key or ""
        value = self.value

        # MUST be a valid hsegment according to RFC 1738, but without * (see PAC-ID Extension)
        # This means it must be true for both, key and value
        if not_allowed_chars := set(re.sub(_hsegment_pattern, '', key)):
            self._add_validation_message(
                    source=f"id segment key {key}",
                    level = ValidationMsgLevel.ERROR,
                    msg=f"{_quote_texts(not_allowed_chars)} must not be used. The segment key must be a valid hsegment",
                    highlight_pattern = key,
                    highlight_sub = not_allowed_chars
            )

        if not_allowed_chars := set(re.sub(_hsegment_pattern, '', value)):
            self._add_validation_message(
                    source=f"id segment key {value}",
                    level = ValidationMsgLevel.ERROR,
                    msg=f"{_quote_texts(not_allowed_chars)} must not be used. The segment key must be a valid hsegment",
                    highlight_pattern = value,
                    highlight_sub = not_allowed_chars
            )

        # Segment key SHOULD be limited to A-Z, 0-9, and -+..
        if not_recommended_chars := set(re.sub(r'[A-Z0-9-:+]', '', key)):
            self._add_validation_message(
                    source=f"id segment key {key}",
                    level = ValidationMsgLevel.RECOMMENDATION,
                    msg=f"{_quote_texts(not_recommended_chars)} should not be used. Characters SHOULD be limited to upper case letters (A-Z), numbers (0-9), '-' and '+' ",
                    highlight_pattern = key,
                    highlight_sub = not_recommended_chars
                )

        # Segment key should be in Well know keys
        if key and key not in [k.value for k in WellKnownKeys]:
            self._add_validation_message(
                    source=f"id segment key {key}",
                    level = ValidationMsgLevel.RECOMMENDATION,
                    msg=f"{key} is not a well known segment key. It is RECOMMENDED to use well-known keys.",
                    highlight_pattern = key,
                    highlight_sub=[key]
                )


        # Segment value SHOULD be limited to A-Z, 0-9, and -+..
        if not_recommended_chars := set(re.sub(r'[A-Z0-9-:+]', '', value)):
            self._add_validation_message(
                    source=f"id segment value {value}",
                    level = ValidationMsgLevel.RECOMMENDATION,
                    msg=f"Characters {_quote_texts(not_recommended_chars)} should not be used., Characters SHOULD be limited to upper case letters (A-Z), numbers (0-9), '-' and '+' ",
                    highlight_pattern = value,
                    highlight_sub = not_recommended_chars
                )

        # Segment value SHOULD be limited to A-Z, 0-9, and :-+ for new designs.
        # this means that ":" in key or value is problematic
        if ':' in key:
            self._add_validation_message(
                    source=f"id segment key {key}",
                    level = ValidationMsgLevel.RECOMMENDATION,
                    msg="Character ':' should not be used in segment key, since this character is used to separate key and value this can lead to undefined behaviour.",
                    highlight_pattern = key
                )
        if ':' in value:
            self._add_validation_message(
                    source=f"id segment value {value}",
                    level = ValidationMsgLevel.RECOMMENDATION,
                    msg="Character ':' should not be used in segment value, since this character is used to separate key and value this can lead to undefined behaviour.",
                    highlight_pattern = value
                )

        return self


from datetime import datetime, timezone


def ensure_utc(dt: datetime) -> datetime:
    '''Converts to UTC time. If dt has no timezone it is assumed to be in UTC and the utc timezone is explicitly added'''
    return dt.replace(tzinfo=timezone.utc) if dt.tzinfo is None else dt.astimezone(timezone.utc)


       
    
from collections import Counter
import logging
import re

from pydantic import RootModel, model_validator
from labfreed.trex.trex_base_models import Value
from labfreed.well_known_keys.unece.unece_units import unece_unit_codes
from labfreed.labfreed_infrastructure import LabFREED_BaseModel, ValidationMsgLevel, _quote_texts
from labfreed.trex.trex_base_models import AlphanumericValue, BinaryValue, BoolValue, DateValue, ErrorValue, NumericValue, TREX_Segment, TextValue


class ColumnHeader(LabFREED_BaseModel):
    '''Header of a table Column'''
    key:str
    type:str
               
    @model_validator(mode='after')     
    def _validate_key(self):
        if not_allowed_chars := set(re.sub(r'[A-Z0-9\.-]', '', self.key)):
            self._add_validation_message(
                source=f"TREX table column {self.key}",
                level= ValidationMsgLevel.ERROR,
                msg=f"Column header key contains invalid characters: {_quote_texts(not_allowed_chars)}",
                highlight_pattern = f'{self.key}$',
                highlight_sub=not_allowed_chars
            )
        return self
    
    @model_validator(mode='after')
    def _validate_type(self):
        valid_types = unece_unit_codes() + ['T.D', 'T.B', 'T.A', 'T.T', 'T.X', 'E']
        if self.type not in valid_types:
            self._add_validation_message(
                    source=f"TREX table column {self.key}",
                    level= ValidationMsgLevel.ERROR,
                    msg=f"Type '{self.type}' is invalid. Must be 'T.D', 'T.B', 'T.A', 'T.T', 'T.X', 'E' or a UNECE unit",
                    highlight_pattern = self.type
            )
        return self

class TableRow(RootModel[list[Value]]):
    """
    Represents a row in a table.

    This class is a Pydantic RootModel that wraps a `list[ValueMixin]`.
    Each element in the list corresponds to a cell in the row.

    All common list operations (indexing, iteration, append, pop, etc.) are supported.
    Internally, it wraps a list in the `.root` attribute.
    """
    def serialize(self):
        return ':'.join([e.serialize() for e in self.root])
    
    def __len__(self):
        return len(self.root)
    
    def __iter__(self):
        return iter(self.root)
    
    def __repr__(self):
        return f"TableRow({self.root!r})  # wraps list[{Value.__name__}]"
    
  
class TableSegment(TREX_Segment):
    '''TREX Segment which represents tabular data'''
    key:str   
    column_headers: list[ColumnHeader]
    data: list[TableRow]
    
    @property
    def column_names(self):
        return [h.key for h in self.column_headers]
    
    @property
    def column_types(self):
        return [h.type for h in self.column_headers]
    
    @model_validator(mode='after')
    def _validate_sizes(self):
        sizes = [len(self.column_headers)]
        sizes.extend( [ len(row) for row in self.data ] ) 
        most_common_len, count = Counter(sizes).most_common(1)[0]        
        
        if len(self.column_headers) != most_common_len:
            self._add_validation_message(
                    source=f"Table {self.key}",
                    level= ValidationMsgLevel.ERROR,
                    msg=f"Size mismatch: Table header contains {self.column_names} keys, while most rows have {most_common_len}",
                    highlight_pattern = self.key
            )
            expected_row_len = most_common_len
        else:
            expected_row_len = len(self.column_headers)
            

        for i, row in enumerate(self.data):
            if len(row) != expected_row_len:
                self._add_validation_message(
                    source=f"Table {self.key}",
                    level= ValidationMsgLevel.ERROR,
                    msg=f"Size mismatch: Table row {i} contains {len(row)} elements. Expected size is {expected_row_len}",
                    highlight_pattern = row.serialize()
                )
        return self
    
    @model_validator(mode='after')
    def _validate_data_types(self):
        expected_types = self.column_types
        i = 0
        for row in self.data:
            for e, t_expected, nm in zip(row, expected_types, self.column_names):
                try:
                    match t_expected:
                        case 'T.D':
                            assert isinstance(e, DateValue)
                        case 'T.B':
                            assert isinstance(e, BoolValue)
                        case 'T.A':
                            assert isinstance(e, AlphanumericValue)              
                        case 'T.T':
                            assert isinstance(e, TextValue)
                        case 'T.X':
                            assert isinstance(e, BinaryValue)
                        case 'E':
                            assert isinstance(e, ErrorValue)
                        case _:
                            assert isinstance(e, NumericValue)
                except AssertionError:
                    self._add_validation_message(
                        source=f"Table {self.key}",
                        level= ValidationMsgLevel.ERROR,
                        msg=f"Type mismatch: Table row {i}, column {nm} is of wrong type. According to the header it should be {t_expected}",
                        highlight_pattern = row.serialize(),
                        highlight_sub=[c for c in e.value]
                    )
                    
                if msg := e.errors():
                    for m in msg:
                        self._add_validation_message(
                            source=f"Table {self.key}",
                            level= ValidationMsgLevel.ERROR,
                            msg=m.msg,
                            highlight_pattern = row.serialize(),
                            highlight_sub=[c for c in e.value]
                        )
                i += 1
        return self
                
            
    def _get_col_index(self, col:str|int):
        if isinstance(col, str):
            col_index = self.column_names.index(col)
        elif isinstance(col, int):
            col_index = col
        else:
            raise TypeError(f"Column must be specified as string or int: {col.__name__}")
        return col_index
    
    
    
    def serialize(self):
        header = ':'.join([f'{h.key}${h.type}' for h in self.column_headers])        
        data = '::'.join([r.serialize() for r in self.data])
        s = f'{self.key}$${header}::{data}'
        return s
    

    def n_rows(self) -> int:
        return len(self.data)
    
    def n_cols(self) -> int:
        return len(self.column_headers)
      
    def row_data(self, row:int) -> list:
        out = self.data[row]
        return out
    
    
    def column_data(self, col:str|int) -> list:
        col_index = self._get_col_index(col)
        out =  [row[col_index] for row in self.data]
        return out
     
     
    def cell_data(self, row:int, col:str|int):
        try:
            col_index = self._get_col_index(col)
            value = self.data[row][col_index]
        except ValueError:
            logging.warning(f"row {row}, column {col} not found")
            return None 
        return value
   



 

def _deserialize_table_segment_from_trex_segment_str(trex_segment_str) -> TableSegment:
    # re_table_pattern = re.compile(f"(?P<tablename>[\w\.-]*?)\$\$(?P<header>[\w\.,\$:]*?)::(?P<body>.*)")
    # re_col_head_pattern = re.compile(f"(?P<name>[\w\.-]*?)\$(?P<unit>[\w\.]*)")
    re_table_pattern = re.compile(r"(?P<tablename>.+?)\$\$(?P<header>.+?)::(?P<body>.+)")
    
    matches = re_table_pattern.match(trex_segment_str) 
    if not matches:
        return None
    name, header, body = matches.groups()
    
    column_headers_str = header.split(':')
    
    headers = []
    for colum_header in column_headers_str:
         ch = colum_header.split('$')
         col_key = ch[0]
         col_type = ch[1] if len(ch) > 1 else ''
         headers.append(ColumnHeader(key=col_key, type=col_type))
    
    data = [row.split(':') for row in body.split('::') ]
    # convert to correct value types
    data_with_types = [[_str_to_value_type(h.type, cv) for cv, h in zip(r, headers)] for r in data]
    data = [ TableRow(r) for r in data_with_types]
             
    out = TableSegment(column_headers=headers, data=data, key=name)
    return out

def _str_to_value_type(type_, s):
    match type_:
        case 'T.D':
            out = DateValue(value=s)
        case 'T.B':
            out = BoolValue(value=s)
        case 'T.A':
            out = AlphanumericValue(value=s)
        case 'T.T':
            out = TextValue(value=s)
        case 'T.X':
            out = BinaryValue(value=s)
        case 'E':
            out = ErrorValue(value=s)
        case _:
            out = NumericValue(value=s)
    return out
        





from datetime import date, datetime, time
from typing import Union
from pydantic import BaseModel, Field, PrivateAttr, model_validator

from labfreed.utilities.base36 import base36
from labfreed.trex.pythonic.quantity import Quantity


class DataTable(BaseModel):
    _row_template:list[str, Quantity | datetime | time | date | bool | str | base36] =  PrivateAttr(default_factory=list)
    col_names: list[str] = Field(default_factory=list)
    data:list[list[Union[Quantity, datetime, time, date, bool, str, base36, None]]] = Field(default_factory=list)
    
    @property
    def row_template(self):
        return self._row_template
    
    @model_validator(mode='after')
    def get_row_template(self):
        if not self.data: # data not initialized during construction. This is valid
            return self
        
        for r in self.data:
            if all([e is not None for e in r]):
                self._row_template = r.copy()
            break
        if not self._row_template:
            raise ValueError('All columns contained at least one None. This is invalid')
        return self
        
       
    def append(self, row:list, validate=True):
        if not isinstance(row, list):
            raise ValueError('row must be a list of values')
        if not self._row_template:
            self._row_template = row.copy()
        if not len(row) == len(self._row_template):
            raise ValueError('row is not of same length as the row template.')
        if not self.col_names:
            self.col_names = [f"Col{i}" for i in range(len(self._row_template))]
        
        # make sure int and float have a unit, if the row_tempalet has one
        for i, e in enumerate(row):
            if isinstance(e, float|int) and isinstance(self._row_template[i], Quantity):
                unit = self._row_template[i].unit
                row[i] = Quantity(value=e, unit=unit)     
        self.data.append(row)
        if validate:
            self.model_rebuild()
        
        
    def extend(self, iterable):
        for item in iterable:
            if not len(item) == len(self._row_template):
                raise ValueError('row is not of same length as the row template.')
            self.append(item, validate=False) 
        self.model_rebuild()
       
            
    def get_column(self, col:str|int) -> list:
        if isinstance(col, str):
            col_index = self.col_names.index(col)
        else:
            col_index = col
        col_data = [row[col_index] for row in self.data]
        return col_data
    
    
    def get_row(self, row_index:int) -> list:
        return self.data[row_index]
    
    
    def get_row_as_dict(self, row_index:int) -> dict:
        d = {k:v for k, v in zip(self.col_names, self.data[row_index])}
        return d
    
    
    def get_cell(self, row_index:int, col:str|int):
        if isinstance(col, str):
            col_index = self.col_names.index(col)
        else:
            col_index = col
        return self.data[row_index][col_index]
        

    


from enum import Enum
import re
from labfreed.labfreed_infrastructure import ValidationMsgLevel, _quote_texts



class ServiceType(Enum):
    USER_HANDOVER_GENERIC = 'userhandover-generic'
    ATTRIBUTE_SERVICE_GENERIC = 'attributes-generic'
    ACTION_GENERIC = "action-generic"


def _validate_service_name(service_name):
    msg_dict = []
    if not_allowed_chars := set(re.sub(r'[A-Za-z0-9\-\x20]', '', service_name)):
        msg_dict.append( {
                "level": ValidationMsgLevel.ERROR,
                "msg": f'Service name ontains invalid characters {_quote_texts(not_allowed_chars)}',
                "highlight_sub": not_allowed_chars
            }
        )
    
    if len(service_name) == 0 or len(service_name) > 255:
        msg_dict.append( {
                "level": ValidationMsgLevel.ERROR,
                "msg": 'Service name must be at least one and maximum 255 characters long'
            }
        )
    return msg_dict
        
    
def _validate_application_intent(intent):
    msg_dict = []
    if re.fullmatch('.*-generic$',  intent):
        msg_dict.append( {
                "level": ValidationMsgLevel.ERROR,
                "msg": "Application intent ends with '-generic'. This is not permitted, since it is reserved for future uses'",
                "highlight_sub": [intent]
            }
        )

    if not_allowed_chars := set(re.sub(r'[A-Za-z0-9\-]', '', intent)):
        msg_dict.append( {
                "level": ValidationMsgLevel.ERROR,
                "msg": f'Application intent contains invalid characters {_quote_texts(not_allowed_chars)}',
                "highlight_sub": not_allowed_chars
            }
        )
        
    if len(intent) == 0 or len(intent) > 255:
        msg_dict.append( {
                "level": ValidationMsgLevel.ERROR,
                "source": f'Application intent  {intent}',
                "msg": 'Must be at least one and maximum 255 characters long'
            }
        )
    return msg_dict

    
def _validate_service_type(service_type):
    msg_dict = []
    if isinstance(service_type, ServiceType):
            service_type= service_type.value
    else:
        service_type= service_type
    allowed_types = [ServiceType.ATTRIBUTE_SERVICE_GENERIC.value, ServiceType.USER_HANDOVER_GENERIC.value, ServiceType.ACTION_GENERIC.value]
    if service_type not in allowed_types:
        msg_dict.append( {
                "level": ValidationMsgLevel.ERROR,
                "msg": f'Invalid service type. Must be {_quote_texts(allowed_types)} must be at least one and maximum 255 characters long',
                "highlight_sub": service_type
            }
        )
    return msg_dict


def _add_msg_to_cit_entry_model(msg_dict, model):
    for m in msg_dict:
        m.update({"source": model.service_name})
        model._add_validation_message(**m)
    return model
    
   


from pydantic import BaseModel, model_validator
from typing import List, Optional
import re


class Translation(BaseModel):
    language_code: str
    text: str
    
    @staticmethod
    def create(language_code:str, text:str):
        '''provides a shorter form to init the model without the need to use keywords only'''
        return Translation(language_code=language_code, text=text)
    

    @model_validator(mode="after")
    def validate_language(self):
        if not re.fullmatch(r"^[a-z]{2,3}(-[a-z]{2,3})?$", self.language_code.lower()):
            raise ValueError(f"Invalid language code: {self.language_code}")
        return self
    
    
    @property
    def language(self) -> str:
        return self.language_code.split("-")[0].lower()

    @property
    def region(self) -> Optional[str]:
        parts = self.language_code.split("-")
        return parts[1].upper() if len(parts) == 2 else None
    
class Term(BaseModel):
    key: str
    translations: List[Translation]
    
    @staticmethod
    def create(key:str, translations:List[Translation]|List[tuple]):
        '''provides a shorter form to init the model without the need to use keywords only'''
        translations = [ t if isinstance(t, Translation) else Translation.create(*t) for t in translations]
        return Term(key=key, translations=translations)
    
    
    def in_language(self, language_key: str) -> str | None:
        for t in self.translations:
            if t.language_code == language_key:
                return t.text
        return None
    
class Terms(BaseModel):
    terms: List[Term]

    def translations_for_term(self, term_key: str) -> Term | None:
        for t in self.terms:
            if t.key == term_key:
                return t
        return None
    

    




import logging
from typing import Literal, Self
from pydantic import  model_validator
from labfreed.labfreed_infrastructure import LabFREED_BaseModel
from labfreed.pac_id.extension import ExtensionBase
from labfreed.well_known_extensions.text_base36_extension import TextBase36Extension

from labfreed.utilities.base36 import from_base36


class DisplayNameExtension(TextBase36Extension, LabFREED_BaseModel):
    name:Literal['N'] = 'N'
    type:Literal['TEXT'] = 'TEXT'
    
    @model_validator(mode='before')
    def move_display_name_to_text(cls, data):
        # if display_name provided, move it to text
        if isinstance(data, dict) and 'display_name' in data:
            data['text'] = data.pop('display_name')
        return data
    
    @staticmethod
    def from_extension(ext:ExtensionBase) -> Self:
        return DisplayNameExtension.create(name=ext.name,
                                  type=ext.type,
                                  data=ext.data)

    @property
    def display_name(self) -> str:
        return self.text 
    
    
    @staticmethod
    def create(*, name, type, data):
        if name != 'N':
            logging.warning(f'Name {name} was given, but this extension should only be used with name "N". Will ignore input')
            
        if type != 'TEXT':
            logging.warning(f'Type {name} was given, but this extension should only be used with type "TEXT". Will try to parse data as display names')
        
        display_name = from_base36(data)
         
        return DisplayNameExtension(display_name=display_name)
    
    def __str__(self):
        return 'Display name: '+ self.display_name




from enum import Enum
from functools import wraps
import json
import logging
from typing import Any, Protocol
from urllib.parse import unquote, unquote_plus

from flask import Blueprint, current_app, make_response, redirect, url_for, send_from_directory
from labfreed.pac_attributes.api_data_models.request import AttributeRequestData
from labfreed.pac_attributes.server.server import AttributeGroupDataSource, AttributeServerRequestHandler, InvalidRequestError, TranslationDataSource

try:
    from flask import Flask, Response, request
except ImportError:
    raise ImportError("Please install labfreed with the [extended] extra: pip install labfreed[extended]")



class Authenticator(Protocol):
    def __call__(self, request) -> bool: ...
    
class NoAuthRequiredAuthenticator(Authenticator):
    def __call__(self, request) -> bool:
        return True

class Webframework(Enum):
    FLASK = "flask"
    FASTAPI = 'fastapi'
    

class AttributeServerFactory():
    @staticmethod
    def create_server_app( datasources:list[AttributeGroupDataSource], 
                           default_language:str,
                           translation_data_sources:list[TranslationDataSource],
                           authenticator: Authenticator|None,
                           framework:Webframework=Webframework.FLASK,
                           doc_text:str=""
                           ):
        
        if not authenticator:
            raise ValueError("authenticator missing. Either define your own authenticator by implementing the 'Authenticator' Protocol, or - if you do not need authentication - explicitly pass a 'NoAuthRequiredAuthenticator' object")
            
        request_handler = AttributeServerRequestHandler(data_sources=datasources, 
                                                        translation_data_sources= translation_data_sources, 
                                                        default_language=default_language
                                                        )
            
        match(framework):
            case Webframework.FLASK:
                app = AttributeFlaskApp(request_handler,authenticator=authenticator, doc_text=doc_text)
                return app
            case Webframework.FASTAPI:
                raise NotImplementedError('FastAPI webapp not implemented')

            


            
class AttributeFlaskApp(Flask):
    def __init__(self, request_handler: AttributeServerRequestHandler, authenticator: Authenticator | None = None, doc_text:str="", **kwargs: Any):
        super().__init__(__name__, **kwargs)
        self.config['ATTRIBUTE_REQUEST_HANDLER'] = request_handler
        self.config['AUTHENTICATOR'] = authenticator
        self.config['DOC_TEXT'] = doc_text
        
        bp = self.create_attribute_blueprint(request_handler, authenticator)
        self.register_blueprint(bp)

    @staticmethod
    def create_attribute_blueprint(
        request_handler: AttributeServerRequestHandler,
        authenticator: Authenticator | None = None,
    ) -> Blueprint:
        bp = Blueprint("attribute", __name__)


        @bp.get("/<path:pac_id_url_encoded>", strict_slashes=False)
        def handle_attribute_request(pac_id_url_encoded):
            
            if authenticator and not authenticator(request):
                return Response(
                    "Unauthorized", 401,
                    {"WWW-Authenticate": 'Basic realm="Login required"'}
                )
            try:
                request_data = AttributeRequestData.from_http_request(pac_id = pac_id_url_encoded,
                                                                      params = request.args,
                                                                     headers = request.headers)
                response_body = request_handler.handle_attribute_request(request_data)
            except InvalidRequestError as e:
                print(e)
                return "Invalid request", 400
            except Exception as e:
                print(e)
                return "The request was valid, but the server encountered an error", 500
            return (response_body, 200, {"Content-Type": "application/json; charset=utf-8"})
        
        
        @bp.post("/", strict_slashes=False)
        def handle_attribute_request_legacy(pac_id):
            if request.method == 'POST':
                return '\n'.join(('POST request was part of the DRAFT specification, but was changed to GET.',
                                  'You are probably using an pre-release version of the python package.',
                                  'update to the newest version "pip install labfreed"'
                                  )
                )
        
        
        @bp.get("/",  strict_slashes=False)
        @bp.get("/capabilities",  strict_slashes=False)
        def capabilities():
            doc_text = current_app.config.get('DOC_TEXT', "") 
            capabilities = request_handler.capabilities()
            authentication_required = bool(current_app.config.get('AUTHENTICATOR'))
            example_request = AttributeRequestData(pac_id='HTTPS://PAC.METTORIUS.COM/EXAMPLE', language_preferences=['fr', 'de']).model_dump_json(indent=2, exclude_none=True, exclude_unset=True)
            server_address = request.url.replace('/capabilities','').rstrip('/')
            css_url = url_for("static", filename="style.css")
            response = f'''
                <html>
                    <head>
                        <link rel="stylesheet" type="text/css" href="{css_url}">
                    </head>

                    <body>
                        This is a <h1>LabFREED attribute server </h1>
                        <h2>Capabilities</h2>
                        Available Attribute Groups <br> {''.join([f'<a href="{ag}"> {ag} </a><br>' for ag in capabilities.available_attribute_groups])} <br>
                        
                        Supported Languages: {', '.join([f'<b> {l} </b>' for l in capabilities.supported_languages])} <br><br>
                        Default Language: <b>{capabilities.default_language}</b> <br>
                        

                        <h2>How to use</h2>
                        Make a <b>GET</b> request to <a href="{server_address}">{server_address}/<url encoded PAC-ID> </a> 
                        <br><br>
                        Query parameters (optional):<br>
                          attr_grps (optional): An comma separated list of attribute group keys. MUST be url-encoded.  <br>
                          attr_fwd_lkp (optional): Boolean flag ('true' or 'false'). Instructs the server to not include attributes of PAC-IDs which are attributes of type reference of the requested PAC-ID. Defaults to true <br>
                        
                        <br>
                        Consult <a href="https://github.com/ApiniLabs/PAC-Attributes"> the specification </a> for details. <br>

                        
                        {'<h2> Authentication </h2> This server <b> requires authentication </b> ' if authentication_required else ''}  
                        <br>
                        
                        {"<h2>Further Information</h2>"if doc_text else ""} 
                        {doc_text or ""} 
                        

                    </body>
                </html>
    '''
        
            return response
        
        @bp.get("/favicon.<ext>", strict_slashes=False)
        def favicon(ext):
            return send_from_directory("static", f"favicon.{ext}")

        return bp
    
    
    
    



    

    
    
    



from __future__ import annotations  # optional in 3.11, but recommended for consistency

from typing import Self
from pydantic import computed_field, model_validator

from rich import print
from rich.text import Text
from rich.table import Table

from labfreed.labfreed_infrastructure import ValidationMsgLevel

from labfreed.pac_cat.category_base import Category
from labfreed.pac_cat.predefined_categories import category_key_to_class_map
from labfreed.pac_id.id_segment import IDSegment
from labfreed.pac_id.pac_id import PAC_ID

''' Configure pdoc'''

    
class PAC_CAT(PAC_ID):
    ''' 
    Extends a PAC-ID with interpretation of the identifier as categories
    '''
    @computed_field
    @property
    def categories(self) -> list[Category]: 
        '''The categories present in the PAC-ID's identifier'''
        category_segments = self._split_segments_by_category(self.identifier)
        categories = list()
        for c in category_segments:
            categories.append(self._cat_from_cat_segments(c))
        return categories
    
    

    def get_category(self, key) -> Category:
        """Helper to get a category by key
        """ 
        tmp = [c for c in self.categories if c.key == key]
        if not tmp:
            return None
        return tmp[0]
    
    
    @classmethod
    def from_categories(cls, issuer:str, categories:list[Category]) -> PAC_CAT:
        identifier = list()
        for category in categories:
            identifier.append(IDSegment(value=category.key))
            identifier.extend(category.segments)
        return PAC_CAT(issuer=issuer, identifier=identifier)
    
    
    @classmethod
    def from_pac_id(cls, pac_id:PAC_ID) -> PAC_CAT:
        '''Constructs a PAC-CAT from a PAC-ID'''
        return PAC_CAT(issuer=pac_id.issuer, identifier=pac_id.identifier)
    

    
    def to_pac_id(self) -> PAC_ID:
        return PAC_ID(issuer=self.issuer, identifier=self.identifier)
        
            
    @classmethod
    def _cat_from_cat_segments(cls, segments:list[IDSegment]) -> Category:
        segments = segments.copy()
        category_key = segments[0].value         
        segments.pop(0)
        
        known_cat = category_key_to_class_map.get(category_key)
        
        if not known_cat:
            return Category(key=category_key, segments=segments)

        # implicit segment keys
        model_dict = {v.alias: None for k, v in known_cat.model_fields.items() if v.alias and k not in ['key','additional_segments']}
        for k, seg in zip(model_dict.keys(), segments.copy()):
            if seg.key:
                break
            model_dict[k] = seg.value
            segments.pop(0)
            
        # try to fill model keys if not already set
        for s in segments.copy():
            if s.key in model_dict.keys() and not model_dict.get(s.key):
                model_dict[s.key] = s.value
                segments.remove(s)
            
        model_dict['additional_segments'] = segments
        model_dict['key'] = category_key
        cat= known_cat(**model_dict)
        return cat
    
    @staticmethod
    def _split_segments_by_category(segments:list[IDSegment]) -> list[list[IDSegment]]:
        category_segments = list()
        c = list()
        for s in segments:
            # new category starts with "-"
            if s.value[0] == '-':
                c = [s]
                category_segments.append(c)
            else:
                c.append(s)
                
        # first cat can be empty > remove
        category_segments = [c for c in category_segments if len(c) > 0]
                            
        return category_segments
    
    
    @model_validator(mode='after')
    def _check_keys_are_unique_in_each_category(self) -> Self:
        for c in self.categories:
            keys = [s.key for s in c.segments if s.key]
            duplicate_keys = [k for k in set(keys) if keys.count(k) > 1]
            if duplicate_keys:
                for k in duplicate_keys:
                    self._add_validation_message(
                        source=f"identifier {k}",
                        level = ValidationMsgLevel.ERROR,
                        msg=f"Duplicate key {k} in category {c.key}",
                        highlight_pattern = k
                    )
        return self
        
        
    @model_validator(mode='after')
    def _check_identifier_segment_keys_are_unique(self) -> Self:
        ''' override the validator of PAC-ID: in PAC-CAT segments can replicate in different categories'''
        return self
    
        
    def print_categories(self):
        table = Table(title=f'Categories in {str(self)}', show_header=False)
        table.add_column('0')
        table.add_column('1')
        for i, c in enumerate(self.categories):
            if i == 0:
                title = Text('Main Category', style='bold')
            else:
                title = Text('Category', style='bold')
            
            table.add_row(title)
            
            for field_name, field_info in c.model_fields.items():
                if not getattr(c, field_name):
                    continue
                table.add_row(f"{field_name} ({field_info.alias or ''})",
                              f" {getattr(c, field_name)}"
                              )      
            table.add_section()        
        print(table)
        

        



from labfreed.pac_attributes.pythonic.py_attributes import pyAttributes
from labfreed.pac_attributes.server.attribute_data_sources import Dict_DataSource


class pyDict_DataSource(Dict_DataSource):
    def __init__(self, data:dict[str, pyAttributes], *args, **kwargs):
        if not all([isinstance(e, pyAttributes) for e in data.values()]):
            raise ValueError('Invalid data')
        d = {k: v.to_payload_attributes() for k,v in data.items()}
        super().__init__(data=d, *args, **kwargs)

        
    


          
from io import StringIO

class StringIOLineBreak(StringIO):
    def __init__(self, *args, markup=None,  **kwargs):
        self._markup = markup
        super().__init__(*args, **kwargs)
    
    def write(self, s:str):
        s = s + '\n'
        super().write(s)
        
    def write_indented(self, s:str):
        s = '    ' + s + '\n'
        super().write(s)
        
    def title1(self, s):
        if self._markup == 'rich':
            s = f'[bold][underline]{s}[/underline][/bold]'
        elif self._markup == 'kivy':
            s = f'[b][u]{s}[/u][/b]'
        elif self._markup == 'html':
            s = f'<h1>{s}</h1>'
        self.new_section()
        self.write(s)  
        
    def title2(self, s):
        if self._markup == 'rich':
            s = f'[bold]{s}[/bold]'
        elif self._markup == 'kivy':
            s = f'[b]{s}[/b]'
        elif self._markup == 'html':
            s = f'<h2>{s}</h2>'
        self.new_paragraph()
        self.write(s) 
        
    def key_value(self, k, v):
        if not k:
            self.write_indented(v)
            return
            
        if self._markup == 'rich':
            s = f'[bold]{k}[/bold]:  {v}'
        elif self._markup == 'kivy':
            s = f'[b]{k}[/b]:  {v}'
        elif self._markup == 'html':
            s = f'<b>{k}</b>:  {v}'
        self.write_indented(s)
        
    def link(self, s, link):
        if self._markup == 'rich':
            s = f'[bold]{s}[/bold]:  [link={link}]{link}[/link] '
        elif self._markup == 'kivy':
            s = f'[b]{s}[/b]:  [ref={link}]{link}[/ref]'
        elif self._markup == 'html':
            s = f'<b>{s}</b>: <a href={link}>{link}</a>' 
        self.write_indented(s)
        
    
    def new_paragraph(self):
        super().write('\n')    
    
    def new_section(self):
        super().write('\n\n')



import logging
from flask import render_template, request
from labfreed.pac_attributes.api_data_models.request import AttributeRequestData
from labfreed.pac_attributes.api_data_models.response import AttributeGroup
from labfreed.pac_cat.pac_cat import PAC_CAT
from labfreed.pac_cat.predefined_categories import Material_Device, Material_Consumable, Material_Substance


from labfreed.pac_attributes.pythonic.py_attributes import pyAttributes, pyAttribute
from labfreed.pac_attributes.server.attribute_data_sources import AttributeGroupDataSource
from labfreed.pac_attributes.server.translation_data_sources import DictTranslationDataSource

from labfreed.pac_attributes.server.server import AttributeServerRequestHandler


from labfreed.pac_attributes.pythonic.py_attributes import pyReference, pyResource
from labfreed.trex.pythonic import DataTable


from labfreed.utilities.translations import Terms, Term


from urllib.parse import urlparse

import requests
from requests import Response


import socket
from functools import lru_cache




def is_device(pac_url):
    pac_cat = PAC_CAT.from_url(pac_url)
    if not isinstance(pac_cat, PAC_CAT):
        return False
    is_device = isinstance(pac_cat.categories[0], Material_Device)
    return is_device

def model(pac_url):
    if not is_device(pac_url):
        return None
    cat:Material_Device = PAC_CAT.from_url(pac_url).categories[0]
    model = cat.model_number
    return model



class DynamicDemoAttributeGroup(AttributeGroupDataSource):
    
    def __init__(self, data, pac_filter_predicate=None, **kwargs):
        self._data = data
        self._pac_filter_predicate = pac_filter_predicate
        super().__init__(**kwargs)
     
    def is_static(self) -> bool:
        return False
    
    @property
    def provides_attributes(self):
        return [ d[0] for d in self._data ] 
    
    @property
    def translations(self) -> DictTranslationDataSource:
        return DictTranslationDataSource(
            supported_languages={'en'},
            data=Terms( terms=[ Term.create(d[0], d[2]) for d in self._data ])
        )

    
    def attributes(self, pac_url: str) -> AttributeGroup:
        if not self._pac_filter_predicate(pac_url):
            return None
        
        attributes = pyAttributes( [pyAttribute(key=d[0], value=d[1]) for d in self._data] ).to_payload_attributes()
        return AttributeGroup(group_key=self._attribute_group_key, 
                              attributes=attributes)
        
        
        
def product_number_from_pac_url(pac_url:str) -> str | None:
    """Safely extract model_number from PAC_CAT"""
    try:
        cat = PAC_CAT.from_url(pac_url)
        if not isinstance(cat, PAC_CAT):
            return None
        main_cat = PAC_CAT.from_url(pac_url).categories[0]
        if isinstance(main_cat, Material_Device):
            main_cat:Material_Device
            return main_cat.model_number
        if isinstance(main_cat, Material_Consumable):
            main_cat:Material_Consumable
            return main_cat.product_number
        if isinstance(main_cat, Material_Substance):
            main_cat:Material_Substance
            return main_cat.product_number
    except Exception:
        return None
    
    
class SessionLocalDirectCall(requests.Session):
    def __init__(self, request_handlers:dict[str, AttributeServerRequestHandler]):
        super().__init__() 
        self._request_handlers = request_handlers
        
    
    def get(self, url, *args, **kwargs):
    
        if is_self_request(url):
            # Case: server calls itself
            # handle differently, e.g., bypass HTTP, call function directly
            logging.warning(f'Handling call to {url} internally')
            parsed_url = urlparse(url)
            return self._handle_direct_call(parsed_url.path, *args, **kwargs)
        else:
            # normal POST via requests
            return super().post(url, *args, **kwargs)

    def _handle_direct_call(self, path:str, *args, **kwargs):
        # Example: directly call the Flask view function instead of HTTP
        # You could map URLs to functions if you know your routing
        # For now, just return a mock response
        path, pac = path.strip("/").rsplit("/", 1)
        rh = self._request_handlers.get(path)
        
        r = Response()
        if rh:
            request_data = AttributeRequestData.from_http_request(pac_id = pac,
                                                                params = kwargs.get('params'),
                                                                 headers = kwargs.get('headers'))
            body = rh.handle_attribute_request(request_data=request_data)
            r.status_code = 200
            r._content = body.encode("utf-8")
            r.encoding = "utf-8"
            r.headers["Content-Type"] = "application/json; charset=utf-8"
        else:
            r.status_code = 404
            r._content = b"Invalid Request"

        return r
        


@lru_cache(maxsize=128)
def resolve_ip(host: str) -> str:
    """Resolve hostname to IP (with caching)."""
    return socket.gethostbyname(host)


def is_self_request(url: str) -> bool:
    parsed = urlparse(url)      
    
    target_host = parsed.hostname.lower()
    current_host = request.host.split(":")[0].lower()
    
    logging.warning(f'Request to target IP: {target_host}, current IP: {current_host}')

    # Shortcut for obvious loopback / link-local
    # Azur case (169.254.*.* → always self).
    if target_host in {"127.0.0.1", "localhost"} or target_host.startswith("169.254."):
        return True

    target_ip = resolve_ip(parsed.hostname.lower())
    current_ip = resolve_ip(request.host.split(":")[0].lower())
    
    return target_ip == current_ip


    
    
    
    
    
def render_template_with_results(pac_id, pac_info=None, cit="", hide_attribute_groups=[], pac_card_url_for=None, route_pac_id_url_for=None):
    return render_template('pac_info_main.jinja.html', 
                pac=pac_id, 
                pac_info = pac_info, 
                cit = cit,
                hide_attribute_groups=hide_attribute_groups, 
                is_data_table = lambda value: isinstance(value, DataTable),
                is_url = lambda s: isinstance(s, str) and urlparse(s).scheme in ('http', 'https') and bool(urlparse(s).netloc),
                is_image = lambda s: isinstance(s, pyResource) and s.root.lower().startswith('http') and s.root.lower().endswith(('.jpg','.jpeg','.png','.gif','.bmp','.webp','.svg','.tif','.tiff')),
                is_reference = lambda s: is_pac_id(s) or isinstance(s, pyReference),
                is_pac_id = is_pac_id,
                pac_card_url_for = pac_card_url_for,
                route_pac_id_url_for = route_pac_id_url_for
                )
    
def is_pac_id(v:str) -> bool:
    try:
        p = PAC_CAT.from_url(v)
        return True and 'PAC.' in v.upper()
    except Exception:
        return False




from dataclasses import dataclass
from functools import cache, partial, partialmethod, wraps
import logging
import os
from pathlib import Path
import re
import secrets
from typing import Any, Protocol
from uuid import uuid4
from jinja2 import ChoiceLoader, FileSystemLoader
import jinja2
from labfreed.labfreed_extended.pac_issuer_lib.lib.utils import add_ga_params, add_trace_id_params
from pydantic import BaseModel, Field

from flask import Blueprint, Flask, Response, current_app, flash, make_response, render_template, request, send_from_directory, session, url_for, make_response
import flask_cors


from werkzeug.middleware.proxy_fix import ProxyFix


from labfreed.pac_cat.pac_cat import PAC_CAT

from labfreed.pac_attributes.pythonic.py_attributes import pyReference, pyResource
from labfreed.trex.pythonic import DataTable

from urllib.parse import urlparse

from labfreed import PAC_ID
from labfreed.labfreed_extended.app.app_infrastructure import Labfreed_App_Infrastructure
from labfreed.pac_attributes.pythonic.attribute_server_factory import AttributeFlaskApp
from labfreed.pac_attributes.server.server import AttributeServerRequestHandler

from labfreed.pac_attributes.server.attribute_data_sources import AttributeGroupDataSource
from labfreed.pac_attributes.server.translation_data_sources import TranslationDataSource, DictTranslationDataSource, Terms

from labfreed.labfreed_extended.app.pac_info.pac_info import PacInfo



from labfreed.labfreed_extended.pac_issuer_lib.lib.attribute import SessionLocalDirectCall

from werkzeug.middleware.proxy_fix import ProxyFix



class NavItem(BaseModel):
    href:str
    name:str
    icon_filename:str

class SiteMeta(BaseModel):
    site_title:str = "LabFREED Starter Kit Page"
    site_description:str = "LabFREED Starter Kit Page"
    site_author:str = "Unknown"
    nav_items:list[NavItem]|None = None
       
    

#TODO: find out why this doesnt work with BaseModel
@dataclass
class AttributeData():
    data_sources:list[AttributeGroupDataSource]
    default_language:str
    translation_data_sources:list[TranslationDataSource]
    
    
    
class PacInfoExtender(Protocol):
    ''' Protocol to inject an object, which analyzes a pac_info and extracts data for issuer specific use cases.'''
    @staticmethod
    def extend(pac_info: PacInfo) -> type[PacInfo]:
        ... 




class IssuerFlaskAppFactory():
    @classmethod
    def create_app( cls, 
                    issuer:str,
                    site_meta:SiteMeta|None = None, 
                    attribute_data:AttributeData|None = None, 
                    path_to_custom_resources:str|None = None,
                    app_secret:str|None = None,
                    pac_info_extender:PacInfoExtender|None = None,
                    resolver_macros:dict[str, str] = None,
                    use_issuer_resolver_config=False,
                    feature_flags:dict|None = None):
        
        app =  Flask(__name__, static_folder=None, static_url_path='/static') 
        
        if not feature_flags:
            feature_flags = dict()
        app.config['feature_flags'] = feature_flags
        
        app.wsgi_app = ProxyFix(app.wsgi_app, x_proto=1, x_host=1)

        
        if not app_secret:
            app.secret_key = str(secrets.token_hex())
        
        bp = cls.create_blueprint( issuer= issuer,
                                    as_multi_issuer_app=False,
                                    site_meta=site_meta,
                                    attribute_data=attribute_data,
                                    path_to_custom_resources=path_to_custom_resources,
                                    pac_info_extender=pac_info_extender,
                                    resolver_macros=resolver_macros,
                                    use_issuer_resolver_config = use_issuer_resolver_config)
        app.register_blueprint(bp)
        
        flask_cors.CORS(app,
            origins="*",               # or a list of allowed origins
            methods=["GET"],           # list of allowed methods
            allow_headers=["Authorization", "Content-Type"],
            supports_credentials=True,  # if you need Authorization or cookies
            resources={
                r"/attributes/*": {},       # pattern-based CORS
                r"/resolver_config.*": {}
            })
        
        return app
    
    
    @classmethod
    def create_blueprint(   cls,
                            issuer:str,
                            as_multi_issuer_app:bool,
                            site_meta:SiteMeta|None = None, 
                            attribute_data:AttributeData|None = None, 
                            path_to_custom_resources:str|None = None,
                            pac_info_extender:PacInfoExtender|None = None,
                            resolver_macros:dict[str, str] = None, 
                            use_issuer_resolver_config=False):
        
        logging.info('initializing Blueprint')
        
        if not resolver_macros:
            resolver_macros = dict()
           
        issuer_name = issuer.split('.')[0]
       
        bp = Blueprint(issuer_name, __name__)
        
        if as_multi_issuer_app:
            bp.url_prefix = f"/{issuer_name}"
        
        # set up static fallback
        custom_static = Path(path_to_custom_resources) / "static"
        default_static = Path(__file__).parent / "static"

        # Override static route with fallback logic
        @bp.get("/static/<path:filename>")
        def static(filename):
            candidate = custom_static / filename
            if candidate.exists():
                return send_from_directory(custom_static, filename)
            return send_from_directory(default_static, filename)
        
        
        @cache
        def resolve_static_image(base_name: str) -> str | None:
            """
            Find the first matching image file for base_name with optional suffix
            (-something or _something) and allowed extensions.
            Checks custom static first, then default static.
            """
            
            base_name = Path(base_name).stem # This removes the last extension ("logo.svg" → "logo").

            exts = ["svg", "png", "jpg", "jpeg", "webp"]
            
            pattern = re.compile(rf"^{re.escape(base_name)}([-_]\w+)?\.({'|'.join(exts)})$")

            # check in custom first
            for folder in (custom_static, default_static):
                if not folder.exists():
                    continue
                for file in folder.iterdir():
                    if file.is_file() and pattern.match(file.name):
                        return url_for_within_issuer(f"static", filename=file.name)

            return None


        def url_for_within_issuer(endpoint, *args, **kwargs):
            if endpoint == "static":
                p = f'{issuer_name}.{endpoint}'
                url = url_for(p, *args, **kwargs)
                return url
            return url_for(endpoint, *args, **kwargs)
        
        
        @bp.context_processor
        def inject_helpers():
            return {"resolve_static_image": resolve_static_image,
                    "url_for_within_issuer": url_for_within_issuer,
                    'url_for': url_for_within_issuer
                    }
            
        def register_helpers(_bp):
            @_bp.context_processor
            def inject_helpers():
                return {"resolve_static_image": resolve_static_image,
                        "url_for_within_issuer": url_for_within_issuer,
                        'url_for': url_for_within_issuer
                        }
                
        register_helpers(bp)
            
        
        @bp.context_processor
        def inject_meta():
            d =   site_meta.model_dump() 
            return d
        
        
        '''
        Set up the attribute server
        =================
        We will create a server app with Flask. 
        '''
        if attribute_data:
            request_handler = AttributeServerRequestHandler(data_sources=attribute_data.data_sources, 
                                                            translation_data_sources= attribute_data.translation_data_sources, 
                                                            default_language=attribute_data.default_language
                                                            )


            bp_attribute_server = AttributeFlaskApp.create_attribute_blueprint(request_handler = request_handler)
            bp_attribute_server.url_prefix = '/attributes'
            register_helpers(bp_attribute_server)
            bp.register_blueprint(bp_attribute_server)



        '''
        Set up the PAC.issuer landing page
        =================
        We will create a server app with Flask. 
        '''
        bp_landing_page = Blueprint('landing_page', __name__ )
        
        
        # set up template loader
        issuer_templates = Path(path_to_custom_resources) / "templates"
        default_templates = Path(__file__).parent / "templates"
        if issuer_templates.exists():
            choice_loader = ChoiceLoader([
                        FileSystemLoader(str(issuer_templates)),
                        FileSystemLoader(str(default_templates))
                        ])
            choice_loader.issuer = issuer_name
            loader = choice_loader
        else:
            loader = FileSystemLoader(str(default_templates))
        bp.jinja_loader = loader
        bp_landing_page.jinja_loader = loader
        bp_attribute_server.jinja_loader = loader
            
            
        if attribute_data:
            if as_multi_issuer_app:
                request_handler_key = f'{issuer_name}/attributes'
            else:
                request_handler_key = 'attributes'
            http_client = SessionLocalDirectCall( request_handlers =
                {
                    request_handler_key: request_handler
                }
            )

            app_infrastructure = Labfreed_App_Infrastructure(language_preferences=attribute_data.default_language, 
                                                             http_client=http_client, 
                                                             use_issuer_resolver_config=use_issuer_resolver_config)
            bp._app_infrastructure = app_infrastructure
        
        
        # in order to have the resolver configuration work out of the box for services on this server 
        # the macro $HOST_URL$ can be used. Before the first request to this blueprint it needs to be initialized with the 
        # host address
        bp._resolver_configuration = None
        bp._resolver_macros = resolver_macros
        @bp.before_app_request
        def init_once():
             # Stays private to this module (since the name is prefixed with _.
            if not bp._resolver_configuration:
                # runs before the first request that hits *any* endpoint,
                # but we make sure it executes only once
                base_url = url_for(f'{issuer_name}.{bp_landing_page.name}.index', _external=True).rstrip('/')
                print("Init with base url", base_url)
                bp._base_url = base_url
                
                resolver_configuration_path = Path(path_to_custom_resources) / 'resolver_configuration.with_macros.yaml'
                if resolver_configuration_path.exists():
                    with open(resolver_configuration_path) as f:
                        rc = f.read()
                        
                    # Replace macros in resolver table
                    _resolver_configuration = rc
                    for macro, subst in bp._resolver_macros.items():
                        old = f'${{{macro}}}'
                        _resolver_configuration = _resolver_configuration.replace(old, subst )
                    _resolver_configuration = _resolver_configuration.replace("${BASE_URL}", base_url )
                    # Replace shorthand for array access by key
                    _resolver_configuration = re.sub(r'\[(".+?")\]', r'[?(@.key == \1)]', _resolver_configuration )
                    
                    logging.info(_resolver_configuration)
                    bp._resolver_configuration = _resolver_configuration
                    
                    app_infrastructure.add_resolver_config(_resolver_configuration)
                    
        # add the pac_info analyzer, too
        bp._pac_info_extender = pac_info_extender
        bp._site_meta = site_meta
        
        
        register_helpers(bp_landing_page)
            
                        
                
        @bp_landing_page.get('/')
        def index():
            return render_from_bp(bp, 'pac_issuer_error.jinja.html'), 404
        


        @bp_landing_page.get('/<path:path>')
        def pac_issuer_landingpage(path):
            if issuer:
                pac_id = f'HTTPS://PAC.{issuer}/{path}'.upper()
            else:
                pac_id = request.url
                pac_id, is_localhost = turn_local_host_to_valid_pac(pac_id, issuer)
            
            try: 
                pac_info = app_infrastructure.process_pac(pac_id)
            except Exception as e:
                logging.exception("Unhandled error")  # full traceback
                render_from_bp(bp, 'pac_issuer_error.jinja.html'), 404
            if not pac_info:
                render_from_bp(bp, 'pac_issuer_error.jinja.html'), 404
            
             
            # let the pac_info extender do it's job, if one was provided
            if a := bp._pac_info_extender:
                a:PacInfoExtender
                pac_info = a.extend(pac_info)
                
            hide_attribute_groups = [] 
            
            pac_card_url_for = partial(url_for,f'{issuer_name}.landing_page.pac_card')
            
            trace_id = uuid4() # used to trace calls to links ( espeially action links)
            session['trace_id']= trace_id
            
            return render_from_bp(
                        bp,
                        "pac_issuer_landing_page.jinja.html",
                        pac=pac_id, 
                        pac_info = pac_info, 
                        hide_attribute_groups=hide_attribute_groups,
                        pac_card_url_for = pac_card_url_for,
                        trace_id = trace_id
                        )
            


        def turn_local_host_to_valid_pac(pac_id, issuer=''):
            if is_localhost := "127.0.0.1" in pac_id:
                #remove localhost from string. including scheme and port, if existing
                pattern = re.compile(r'(?:https?:\/{0,2})?127\.0\.0\.1(?::\d+)?')
                s = re.sub(pattern, '', pac_id)
                if issuer:
                    cleaned_pac_id = f'HTTPS://PAC.{issuer.upper()}/{s}' 
                else:
                    cleaned_pac_id = f'HTTPS://PAC.{s}'
                cleaned_pac_id = cleaned_pac_id
                return cleaned_pac_id, is_localhost
            else:
                return pac_id, is_localhost

            
        @bp_landing_page.get('/resolver_config')
        @bp_landing_page.get('/resolver_config.yaml')
        def resolver_configuration():
            return Response(bp._resolver_configuration, mimetype="application/x-yaml")
        
       
        
        @bp_landing_page.get('/info_card')
        def pac_card():
            pac_id = request.args.get('pac_id')                
            try: 
                p = PAC_ID.from_url(pac_id)
                pac_valid = True
            except Exception as e:
                print(e)
                pac_valid = False
                
            pac_id, is_localhost = turn_local_host_to_valid_pac(pac_id)
            
            if not (pac_valid or is_localhost):
                issuer
                return ""
            
            pac_info = app_infrastructure.process_pac(pac_id)
            if a := bp._pac_info_extender:
                a:PacInfoExtender
                pac_info = a.extend(pac_info)
            response = render_from_bp(bp, 'pac_info/card.jinja.html', 
                                      pac_info=pac_info,
                                      **request.args
                                      )  
            return response
        
        # on local host for testing we need to circumvent the error, which is cause by PAC-ID validation when encountering the local host instead of issuer. 
        # TODO: make this cleaner and more stable
        bp.register_blueprint(bp_landing_page)
        
        @bp.get('/route_pac_id')
        def route_pac_id():
            pac_id = request.args.get('pac_id')
            raise NotImplementedError()
        
                
        def render_from_bp(bp, template_name, **context):
            # Make a new Jinja environment for this bp
            env = jinja2.Environment(loader=bp.jinja_loader)

            # Copy Flask globals/context
            env.globals.update(current_app.jinja_env.globals)
            
            trace_id = session.get('trace_id')
            d = {            
                "url_for_within_issuer": url_for_within_issuer,
                "resolve_static_image": resolve_static_image,
                "url_for": url_for_within_issuer,
                "with_ga_and_trace": lambda url: add_trace_id_params(add_ga_params(url, issuer), trace_id)
            }
            env.globals.update(d)
            
            env.globals.update({'site_meta': bp._site_meta.model_dump()})
            
            env.globals.update(render_context_utils)

            tpl = env.get_template(template_name)
            
            
            
            return tpl.render(**context)
        
        
        
        @bp.errorhandler(Exception)
        def not_found(error:Exception):
            logging.error(msg=error)
            return render_from_bp(bp, "pac_issuer_error.jinja.html", msg=str(error)), 404

        # @bp.errorhandler(500)
        # def internal_error(error):
        #     return render_from_bp("errors/pac_info_500.html"), 500
        
        
            
            
        return bp
    


def is_pac_id(v:str) -> bool:
    try:
        p = PAC_CAT.from_url(v)
        return True and 'PAC.' in v.upper()
    except Exception:
        return False
    
def title_format(s:str) -> str:
    """Convert snake_case string to NYTimes-style title."""
    return s.replace('_', ' ').title()
    
render_context_utils = {            
            "is_url": lambda s: isinstance(s, str) and urlparse(s).scheme in ('http', 'https') and bool(urlparse(s).netloc),
            "is_data_table": lambda value: isinstance(value, DataTable),
            "is_image": lambda s: isinstance(s, pyResource) and s.root.lower().startswith('http') and s.root.lower().endswith(('.jpg','.jpeg','.png','.gif','.bmp','.webp','.svg','.tif','.tiff')),
            "is_reference":lambda s: is_pac_id(s) or isinstance(s, pyReference),
            "is_pac_id": is_pac_id,
            "title_format": title_format,
            "zip": zip
        }
    
    
    
    
    
def attribute_data_from_module(module, default_language):
    try:
        ds = module.data_sources
    except AttributeError as e:
        raise AttributeError(f"Module {module!r} does not define 'data_sources'") from e
        
    try:
        tds = module.translation_data_sources
        if not tds:
            tds = [ DictTranslationDataSource(supported_languages={default_language}, data=Terms(terms=[]))]
    except AttributeError as e:
        tds = [ DictTranslationDataSource(supported_languages={default_language}, data=Terms(terms=[]))]
           
    attribute_data = AttributeData(default_language= default_language, 
                                   data_sources=ds,
                                   translation_data_sources=tds)
    
    return attribute_data






from concurrent.futures import ThreadPoolExecutor, as_completed

import requests

from labfreed.labfreed_extended.app.pac_info.pac_info import PacInfo
from labfreed.pac_attributes.client.client import AttributeClient, http_attribute_request_default_callback_factory
from labfreed.pac_attributes.pythonic.py_attributes import pyAttributeGroup

from labfreed.pac_id.pac_id import PAC_ID
from labfreed.pac_id_resolver.resolver import PAC_ID_Resolver, cit_from_str
from labfreed.pac_id_resolver.services import ServiceGroup

        

class Labfreed_App_Infrastructure():
    def __init__(self, markup = 'rich', language_preferences:list[str]|str='en', http_client:requests.Session|None=None, use_issuer_resolver_config=True):
        if isinstance(language_preferences, str):
            language_preferences = [language_preferences]
        self._language_preferences = language_preferences
        self._use_issuer_resolver_config = use_issuer_resolver_config
        
        self._resolver = PAC_ID_Resolver()
        
        if not http_client:
            http_client = requests.Session()
        self._http_client= http_client
        callback = http_attribute_request_default_callback_factory(http_client)
            
        self._attribute_client = AttributeClient(http_post_callback=callback)


    def add_resolver_config(self, cit:str):
        cit = cit_from_str(cit)
        if not cit:
            raise ValueError('the cit could not be parsed. Neither as v1 or v2')
        self._resolver._resolver_configs.add(cit)
        
    def remove_resolver_config(self, resolver_config:str):
        resolver_config = cit_from_str(resolver_config)
        self._resolver._resolver_configs.discard(resolver_config)
        
        
    def process_pac(self, pac_url) -> PacInfo:
        if not isinstance(pac_url, PAC_ID):
            pac = PAC_ID.from_url(pac_url)
        else:
            pac = pac_url
        service_groups = self._resolver.resolve(pac, check_service_status=False, use_issuer_resolver_config=self._use_issuer_resolver_config)
        
        pac_info = PacInfo(pac_id=pac)
                       
        # update service states
        (sg.update_states() for sg in service_groups)
               
        # Services
        sg_user_handovers = []
        for sg in service_groups:
            user_handovers = [s  for s in sg.services if s.service_type == 'userhandover-generic']
            
            if user_handovers:
                sg_user_handovers.append(ServiceGroup(origin=sg.origin, services=user_handovers))
        pac_info.user_handovers = sg_user_handovers
        
        # Actions
        sg_actions = []
        for sg in service_groups:
            actions = [s  for s in sg.services if s.service_type == 'action-generic']
            
            if actions:
                sg_actions.append(ServiceGroup(origin=sg.origin, services=actions))
        pac_info.actions = sg_actions
        
        # Attributes
        attribute_groups = {}
        for sg in service_groups:  
            attributes_urls = [s.url  for s in sg.services if s.service_type == 'attributes-generic']
            for url in attributes_urls:
                ags = {ag.group_key: pyAttributeGroup.from_attribute_group(ag) for ag in self._attribute_client.get_attributes(url, pac_id=pac.to_url(include_extensions=False), language_preferences=self._language_preferences)}
                if ags:
                    attribute_groups.update(ags)
        pac_info.attribute_groups = attribute_groups
       
        return pac_info
    
    
    
    def update_user_handover_states(self, services, session:requests.Session = None):
        '''Triggers each service to check if the url can be reached'''
        if not _has_internet_connection():
            raise ConnectionError("No Internet Connection")
        with ThreadPoolExecutor(max_workers=10) as executor:
            futures = [executor.submit(s.check_service_status, session=session) for s in services]
            for _ in as_completed(futures):
                pass  # just wait for all to finish
            
            
def _has_internet_connection():
    try:
        requests.head("https://1.1.1.1", timeout=3)
        return True
    except requests.RequestException:
        return False
                
                